﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using ScrobbleMapper.LastFm;
using ScrobbleMapper.Library;
using ScrobbleMapper.Properties;

namespace ScrobbleMapper.Forms
{
    partial class MainForm : Form
    {
        readonly WmpLibraryManager wmpLibrary;
        readonly ItunesLibraryManager iTunesLibrary;

        readonly ScrobbleFetcher fetcher;

        public IEnumerable<ScrobbledTrack> scrobbles;

        public MainForm()
        {
            InitializeComponent();

            UsernameText.Text = Settings.Default.LastFmUsername;
            
            wmpLibrary = new WmpLibraryManager();
            iTunesLibrary = new ItunesLibraryManager();

            fetcher = new ScrobbleFetcher();
        }

        #region Form Events

        void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            Settings.Default.Save();
        }

        void UsernameText_TextChanged(object sender, EventArgs eventArgs)
        {
            FetchButton.Enabled = UsernameText.Text != string.Empty;
        }

        void FetchButton_Click(object sender, EventArgs eventArgs)
        {
            Settings.Default.LastFmUsername = UsernameText.Text;

            TaskUtil.PerformForegroundTask(this, fetcher.FetchAsync(UsernameText.Text), result =>
            {
                scrobbles = result.Scrobbles;
                TracksView.DataSource = new ScrobbledTrackBindingList(scrobbles);
                MapToWmpButton.Enabled = true;
                MapToITunesButton.Enabled = true;
            }, 
                error => MessageBox.Show("This user does not exist, or there was an error fetching his scrobbles." + 
                                         Environment.NewLine + Environment.NewLine +
                                         "Details : " + error.Message, "No results.",
                                         MessageBoxButtons.OK, MessageBoxIcon.Exclamation));
        }

        void MapToWmpButton_Click(object sender, EventArgs eventArgs)
        {
            MapToLibrary(wmpLibrary);
        }

        void MapToITunesButton_Click(object sender, EventArgs eventArgs)
        {
            MapToLibrary(iTunesLibrary);
        }

        #endregion

        #region Actions

        void MapToLibrary(LibraryManager library)
        {
            TaskUtil.PerformForegroundTask(this, library.MapAsync(scrobbles),
                result => ChooseFuzzyMatches(library, result),
                error => MessageBox.Show("An error occured when updating the library." +
                                         Environment.NewLine + Environment.NewLine +
                                         "Details : " + error.Message, "Error!",
                                         MessageBoxButtons.OK, MessageBoxIcon.Error));
        }

        void ChooseFuzzyMatches(LibraryManager library, MapResult result)
        {
            if (result != null && result.FuzzyMatches.Count() > 0)
            {
                TaskUtil.PerformForegroundTask(this, library.ChooseFuzzyMatchesAsync(this, result.FuzzyMatches),
                                fuzzyResult => ShowSummary(result, fuzzyResult),
                                error => MessageBox.Show("An error occured when updating the library." +
                                                         Environment.NewLine + Environment.NewLine +
                                                         "Details : " + error.Message, "Error!",
                                                         MessageBoxButtons.OK, MessageBoxIcon.Error));
            }
        }

        void ShowSummary(MapResult mapResult, ChooseFuzzyMatchesResult fuzzyResult)
        {
            var totalUpdated = mapResult.Updated + fuzzyResult.Updated;
            var totalUpToDate = mapResult.AlreadyUpToDate + fuzzyResult.AlreadyUpToDate;
            var totalFailed = mapResult.UpdateFailed + fuzzyResult.UpdateFailed;
            var totalSkipped = totalFailed + mapResult.NotFound;

            if (totalFailed > 0)
                new ErrorReporter(mapResult.Errors.Concat(fuzzyResult.Errors)).ShowDialog(this);

            MessageBox.Show(totalUpdated + " entries updated, " +
                            totalUpToDate + " were already up to date, " +
                            totalSkipped + " were skipped.",
                            "Mapping complete", MessageBoxButtons.OK,
                            MessageBoxIcon.Information);
        }

        #endregion

        #region Main Menu Events

        void AboutMenuItem_Click(object sender, EventArgs e)
        {
            new AboutBox().ShowDialog();
        }

        void ExitMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        void OptionsMenuItem_Click(object sender, EventArgs e)
        {
            using (var optionsDialog = new OptionsDialog { MinimumEditDistance = Settings.Default.MinimumEditDistance })
            {
                if (optionsDialog.ShowDialog(this) == DialogResult.OK)
                    Settings.Default.MinimumEditDistance = optionsDialog.MinimumEditDistance;
            }
        }

        #endregion
    }
}